import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Any, Dict, Iterator, List, Tuple, Union

import casadi as ca

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.debug_check_helpers import DebugLevel, debug_check
from rtctools.data.storage import DataStoreAccessor

from .timeseries import Timeseries

logger = logging.getLogger("rtctools")


# Typical type for a bound on a variable
BT = Union[float, np.ndarray, Timeseries]


class LookupTable(metaclass=ABCMeta):
    """
    Base class for LookupTables.
    """

    @property
    def inputs(self) -> List[ca.MX]:
        """
        List of lookup table input variables.
        """
        raise NotImplementedError

    @property
    def function(self) -> ca.Function:
        """
        Lookup table CasADi :class:`Function`.
        """
        raise NotImplementedError


class OptimizationProblem(DataStoreAccessor, metaclass=ABCMeta):
    """
    Base class for all optimization problems.
    """

    _debug_check_level = DebugLevel.MEDIUM
    _debug_check_options = {}

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        self.__mixed_integer = False

    def optimize(self, preprocessing: bool = True, postprocessing: bool = True,
                 log_solver_failure_as_error: bool = True) -> bool:
        """
        Perform one initialize-transcribe-solve-finalize cycle.

        :param preprocessing:  True to enable a call to ``pre`` preceding the optimization.
        :param postprocessing: True to enable a call to ``post`` following the optimization.

        :returns: True on success.
        """

        # Deprecations / removals
        if hasattr(self, 'initial_state'):
            raise RuntimeError("Support for `initial_state()` has been removed. Please use `history()` instead.")

        logger.info("Entering optimize()")

        # Do any preprocessing, which may include changing parameter values on
        # the model
        if preprocessing:
            self.pre()

            # Check if control inputs are bounded
            self.__check_bounds_control_input()
        else:
            logger.debug(
                'Skipping Preprocessing in OptimizationProblem.optimize()')

        # Transcribe problem
        discrete, lbx, ubx, lbg, ubg, x0, nlp = self.transcribe()

        # Create an NLP solver
        logger.debug("Collecting solver options")

        self.__mixed_integer = np.any(discrete)
        options = {}
        options.update(self.solver_options())  # Create a copy

        logger.debug("Creating solver")

        if options.pop('expand', False):
            # NOTE: CasADi only supports the "expand" option for nlpsol. To
            # also be able to expand with e.g. qpsol, we do the expansion
            # ourselves here.
            logger.debug("Expanding objective and constraints to SX")

            expand_f_g = ca.Function('f_g', [nlp['x']], [nlp['f'], nlp['g']]).expand()
            X_sx = ca.SX.sym('X', *nlp['x'].shape)
            f_sx, g_sx = expand_f_g(X_sx)

            nlp['f'] = f_sx
            nlp['g'] = g_sx
            nlp['x'] = X_sx

        # Debug check for non-linearity in constraints
        self.__debug_check_linearity_constraints(nlp)

        # Debug check for linear independence of the constraints
        self.__debug_check_linear_independence(lbx, ubx, lbg, ubg, nlp)

        # Solver option
        my_solver = options['solver']
        del options['solver']

        # Already consumed
        del options['optimized_num_dir']

        # Iteration callback
        iteration_callback = options.pop('iteration_callback', None)

        # CasADi solver to use
        casadi_solver = options.pop('casadi_solver')
        if isinstance(casadi_solver, str):
            casadi_solver = getattr(ca, casadi_solver)

        nlpsol_options = {**options}

        if self.__mixed_integer:
            nlpsol_options['discrete'] = discrete
        if iteration_callback:
            nlpsol_options['iteration_callback'] = iteration_callback

        # Remove ipopt and bonmin defaults if they are not used
        if my_solver != 'ipopt':
            nlpsol_options.pop('ipopt', None)
        if my_solver != 'bonmin':
            nlpsol_options.pop('bonmin', None)

        solver = casadi_solver('nlp', my_solver, nlp, nlpsol_options)

        # Solve NLP
        logger.info("Calling solver")

        results = solver(x0=x0, lbx=lbx, ubx=ubx, lbg=ca.veccat(*lbg), ubg=ca.veccat(*ubg))

        # Extract relevant stats
        self.__objective_value = float(results['f'])
        self.__solver_output = np.array(results['x']).ravel()
        self.__solver_stats = solver.stats()

        success, log_level = self.solver_success(self.__solver_stats, log_solver_failure_as_error)

        if success:
            logger.log(log_level, "Solver succeeded with status {}".format(
                self.__solver_stats['return_status']))
        else:
            logger.log(log_level, "Solver failed with status {}".format(
                self.__solver_stats['return_status']))

        # Do any postprocessing
        if postprocessing:
            self.post()
        else:
            logger.debug(
                'Skipping Postprocessing in OptimizationProblem.optimize()')

        # Done
        logger.info("Done with optimize()")

        return success

    def __check_bounds_control_input(self) -> None:
        # Checks if at the control inputs have bounds, log warning when a control input is not bounded.
        bounds = self.bounds()

        for variable in self.dae_variables['control_inputs']:
            variable = variable.name()
            if variable not in bounds:
                logger.warning(
                    "OptimizationProblem: control input {} has no bounds.".format(variable))

    @abstractmethod
    def transcribe(self) -> Tuple[
            np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, Dict[str, ca.MX]]:
        """
        Transcribe the continuous optimization problem to a discretized, solver-ready
        optimization problem.
        """
        pass

    def solver_options(self) -> Dict[str, Union[str, int, float, bool, str]]:
        """
        Returns a dictionary of CasADi optimization problem solver options.

        The default solver for continuous problems is `Ipopt <https://projects.coin-or.org/Ipopt/>`_.
        The default solver for mixed integer problems is `Bonmin <http://projects.coin-or.org/Bonmin/>`_.

        :returns: A dictionary of solver options. See the CasADi and
                  respective solver documentation for details.
        """
        options = {'error_on_fail': False,
                   'optimized_num_dir': 3,
                   'casadi_solver': ca.nlpsol}

        if self.__mixed_integer:
            options['solver'] = 'bonmin'

            bonmin_options = options['bonmin'] = {}
            bonmin_options['algorithm'] = 'B-BB'
            bonmin_options['nlp_solver'] = 'Ipopt'
            bonmin_options['nlp_log_level'] = 2
            bonmin_options['linear_solver'] = 'mumps'
        else:
            options['solver'] = 'ipopt'

            ipopt_options = options['ipopt'] = {}
            ipopt_options['linear_solver'] = 'mumps'
        return options

    def solver_success(self,
                       solver_stats: Dict[str, Union[str, bool]],
                       log_solver_failure_as_error: bool) -> Tuple[bool, int]:
        """
        Translates the returned solver statistics into a boolean and log level
        to indicate whether the solve was succesful, and how to log it.

        :param solver_stats: Dictionary containing information about the
                             solver status. See explanation below.
        :param log_solver_failure_as_error: Indicates whether a solve failure
            Should be logged as an error or info message.

        ``solver_stats`` typically consist of three fields:

        * return_status: ``str``
        * secondary_return_status: ``str``
        * success: ``bool``

        By default we rely on CasADi's interpretation of the return_status
        (and secondary status) to the success variable, with an exception for
        IPOPT (see below).

        The logging level is typically ``logging.INFO`` for success, and
        ``logging.ERROR`` for failure. Only for IPOPT an exception is made for
        `Not_Enough_Degrees_Of_Freedom`, which returns ``logging.WARNING`` instead.
        For example, this can happen when too many goals are specified, and
        lower priority goals cannot improve further on the current result.

        :returns: A tuple indicating whether or not the solver has succeeded, and what level to log it with.
        """
        success = solver_stats['success']
        log_level = logging.INFO if success else logging.ERROR

        if (self.solver_options()['solver'].lower() in ['bonmin', 'ipopt']
                and solver_stats['return_status'] in ['Not_Enough_Degrees_Of_Freedom']):
            log_level = logging.WARNING

        if log_level == logging.ERROR and not log_solver_failure_as_error:
            log_level = logging.INFO

        return success, log_level

    @abstractproperty
    def solver_input(self) -> ca.MX:
        """
        The symbolic input to the NLP solver.
        """
        pass

    @abstractmethod
    def extract_results(self, ensemble_member: int = 0) -> Dict[str, np.ndarray]:
        """
        Extracts state and control input time series from optimizer results.

        :returns: A dictionary of result time series.
        """
        pass

    @property
    def objective_value(self) -> float:
        """
        The last obtained objective function value.
        """
        return self.__objective_value

    @property
    def solver_output(self) -> np.ndarray:
        """
        The raw output from the last NLP solver run.
        """
        return self.__solver_output

    @property
    def solver_stats(self) -> Dict[str, Any]:
        """
        The stats from the last NLP solver run.
        """
        return self.__solver_stats

    def pre(self) -> None:
        """
        Preprocessing logic is performed here.
        """
        pass

    @abstractproperty
    def dae_residual(self) -> ca.MX:
        """
        Symbolic DAE residual of the model.
        """
        pass

    @abstractproperty
    def dae_variables(self) -> Dict[str, List[ca.MX]]:
        """
        Dictionary of symbolic variables for the DAE residual.
        """
        pass

    @property
    def path_variables(self) -> List[ca.MX]:
        """
        List of additional, time-dependent optimization variables, not covered by the DAE model.
        """
        return []

    @abstractmethod
    def variable(self, variable: str) -> ca.MX:
        """
        Returns an :class:`MX` symbol for the given variable.

        :param variable: Variable name.

        :returns: The associated CasADi :class:`MX` symbol.
        """
        raise NotImplementedError

    @property
    def extra_variables(self) -> List[ca.MX]:
        """
        List of additional, time-independent optimization variables, not covered by the DAE model.
        """
        return []

    @property
    def output_variables(self) -> List[ca.MX]:
        """
        List of variables that the user requests to be included in the output files.
        """
        return []

    def delayed_feedback(self) -> List[Tuple[str, str, float]]:
        """
        Returns the delayed feedback mappings.  These are given as a list of triples :math:`(x, y, \\tau)`,
        to indicate that :math:`y = x(t - \\tau)`.

        :returns: A list of triples.

        Example::

            def delayed_feedback(self):
                fb1 = ['x', 'y', 0.1]
                fb2 = ['x', 'z', 0.2]
                return [fb1, fb2]

        """
        return []

    @property
    def ensemble_size(self) -> int:
        """
        The number of ensemble members.
        """
        return 1

    def ensemble_member_probability(self, ensemble_member: int) -> float:
        """
        The probability of an ensemble member occurring.

        :param ensemble_member: The ensemble member index.

        :returns: The probability of an ensemble member occurring.

        :raises: IndexError
        """
        return 1.0

    def parameters(self, ensemble_member: int) -> AliasDict[str, Union[bool, int, float, ca.MX]]:
        """
        Returns a dictionary of parameters.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of parameter names and values.
        """
        return AliasDict(self.alias_relation)

    def string_parameters(self, ensemble_member: int) -> Dict[str, str]:
        """
        Returns a dictionary of string parameters.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of string parameter names and values.
        """
        return {}

    def constant_inputs(self, ensemble_member: int) -> AliasDict[str, Timeseries]:
        """
        Returns a dictionary of constant inputs.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of constant input names and time series.
        """
        return AliasDict(self.alias_relation)

    def lookup_tables(self, ensemble_member: int) -> AliasDict[str, LookupTable]:
        """
        Returns a dictionary of lookup tables.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of variable names and lookup tables.
        """
        return AliasDict(self.alias_relation)

    @staticmethod
    def merge_bounds(a: Tuple[BT, BT], b: Tuple[BT, BT]) -> Tuple[BT, BT]:
        """
        Returns a pair of bounds which is the intersection of the two pairs of
        bounds given as input.

        :param a: First pair ``(upper, lower)`` bounds
        :param b: Second pair ``(upper, lower)`` bounds

        :returns: A pair of ``(upper, lower)`` bounds which is the
                  intersection of the two input bounds.
        """
        a, A = a
        b, B = b

        # Make sure we are dealing with the correct types
        if __debug__:
            for v in (a, A, b, B):
                if isinstance(v, np.ndarray):
                    assert v.ndim == 1
                    assert np.issubdtype(v.dtype, np.number)
                else:
                    assert isinstance(v, (float, int, Timeseries))

        all_bounds = [a, A, b, B]

        # First make sure that we treat single element vectors as scalars
        for i, v in enumerate(all_bounds):
            if isinstance(v, np.ndarray) and np.prod(v.shape) == 1:
                all_bounds[i] = v.item()

        # Upcast lower bounds to be of equal type, and upper bounds as well.
        for i, j in [(0, 2), (2, 0), (1, 3), (3, 1)]:
            v1 = all_bounds[i]
            v2 = all_bounds[j]

            # We only check for v1 being of a "smaller" type than v2, as we
            # know we will encounter the reverse as well.
            if isinstance(v1, type(v2)):
                # Same type, nothing to do.
                continue
            elif isinstance(v1, (int, float)) and isinstance(v2, Timeseries):
                all_bounds[i] = Timeseries(v2.times, np.full_like(v2.values, v1))
            elif isinstance(v1, np.ndarray) and isinstance(v2, Timeseries):
                if v2.values.ndim != 2 or len(v1) != v2.values.shape[1]:
                    raise Exception(
                        "Mismatching vector size when upcasting to Timeseries, {} vs. {}.".format(v1, v2))
                all_bounds[i] = Timeseries(v2.times, np.broadcast_to(v1, v2.values.shape))
            elif isinstance(v1, (int, float)) and isinstance(v2, np.ndarray):
                all_bounds[i] = np.full_like(v2, v1)

        a, A, b, B = all_bounds

        assert isinstance(a, type(b))
        assert isinstance(A, type(B))

        # Merge the bounds
        m, M = None, None

        if isinstance(a, np.ndarray):
            if not a.shape == b.shape:
                raise Exception("Cannot merge vector minimum bounds of non-equal size")
            m = np.maximum(a, b)
        elif isinstance(a, Timeseries):
            if len(a.times) != len(b.times):
                raise Exception("Cannot merge Timeseries minimum bounds with different lengths")
            elif not np.all(a.times == b.times):
                raise Exception("Cannot merge Timeseries minimum bounds with non-equal times")
            elif not a.values.shape == b.values.shape:
                raise Exception("Cannot merge vector Timeseries minimum bounds of non-equal size")
            m = Timeseries(a.times, np.maximum(a.values, b.values))
        else:
            m = max(a, b)

        if isinstance(A, np.ndarray):
            if not A.shape == B.shape:
                raise Exception("Cannot merge vector maximum bounds of non-equal size")
            M = np.minimum(A, B)
        elif isinstance(A, Timeseries):
            if len(A.times) != len(B.times):
                raise Exception("Cannot merge Timeseries maximum bounds with different lengths")
            elif not np.all(A.times == B.times):
                raise Exception("Cannot merge Timeseries maximum bounds with non-equal times")
            elif not A.values.shape == B.values.shape:
                raise Exception("Cannot merge vector Timeseries maximum bounds of non-equal size")
            M = Timeseries(A.times, np.minimum(A.values, B.values))
        else:
            M = min(A, B)

        return m, M

    def bounds(self) -> AliasDict[str, Tuple[BT, BT]]:
        """
        Returns variable bounds as a dictionary mapping variable names to a pair of bounds.
        A bound may be a constant, or a time series.

        :returns: A dictionary of variable names and ``(upper, lower)`` bound pairs.
                  The bounds may be numbers or :class:`.Timeseries` objects.

        Example::

            def bounds(self):
                return {'x': (1.0, 2.0), 'y': (2.0, 3.0)}

        """
        return AliasDict(self.alias_relation)

    def history(self, ensemble_member: int) -> AliasDict[str, Timeseries]:
        """
        Returns the state history.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of variable names and historical time series (up to and including t0).
        """
        return AliasDict(self.alias_relation)

    def variable_is_discrete(self, variable: str) -> bool:
        """
        Returns ``True`` if the provided variable is discrete.

        :param variable: Variable name.

        :returns: ``True`` if variable is discrete (integer).
        """
        return False

    def variable_nominal(self, variable: str) -> Union[float, np.ndarray]:
        """
        Returns the nominal value of the variable.  Variables are scaled by replacing them with
        their nominal value multiplied by the new variable.

        :param variable: Variable name.

        :returns: The nominal value of the variable.
        """
        return 1

    @property
    def initial_time(self) -> float:
        """
        The initial time in seconds.
        """
        return self.times()[0]

    @property
    def initial_residual(self) -> ca.MX:
        """
        The initial equation residual.

        Initial equations are used to find consistent initial conditions.

        :returns: An :class:`MX` object representing F in the initial equation F = 0.
        """
        return ca.MX(0)

    def seed(self, ensemble_member: int) -> AliasDict[str, Union[float, Timeseries]]:
        """
        Seeding data.  The optimization algorithm is seeded with the data returned by this method.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of variable names and seed time series.
        """
        return AliasDict(self.alias_relation)

    def objective(self, ensemble_member: int) -> ca.MX:
        """
        The objective function for the given ensemble member.

        Call :func:`OptimizationProblem.state_at` to return a symbol representing a model variable at a given time.

        :param ensemble_member: The ensemble member index.

        :returns: An :class:`MX` object representing the objective function.

        Example::

            def objective(self, ensemble_member):
                # Return value of state 'x' at final time:
                times = self.times()
                return self.state_at('x', times[-1], ensemble_member)

        """
        return ca.MX(0)

    def path_objective(self, ensemble_member: int) -> ca.MX:
        """
        Returns a path objective the given ensemble member.

        Path objectives apply to all times and ensemble members simultaneously.
        Call :func:`OptimizationProblem.state` to return a time- and ensemble-member-independent
        symbol representing a model variable.

        :param ensemble_member: The ensemble member index. This index is currently unused,
                                and here for future use only.

        :returns: A :class:`MX` object representing the path objective.

        Example::

            def path_objective(self, ensemble_member):
                # Minimize x(t) for all t
                return self.state('x')

        """
        return ca.MX(0)

    def constraints(self, ensemble_member: int) -> List[
            Tuple[ca.MX, Union[float, np.ndarray], Union[float, np.ndarray]]]:
        """
        Returns a list of constraints for the given ensemble member.

        Call :func:`OptimizationProblem.state_at` to return a symbol representing a model variable at a given time.

        :param ensemble_member: The ensemble member index.

        :returns: A list of triples ``(f, m, M)``, with an :class:`MX` object representing
                  the constraint function ``f``, lower bound ``m``, and upper bound ``M``.
                  The bounds must be numbers.

        Example::

            def constraints(self, ensemble_member):
                t = 1.0
                constraint1 = (
                    2 * self.state_at('x', t, ensemble_member),
                    2.0, 4.0)
                constraint2 = (
                    self.state_at('x', t, ensemble_member) + self.state_at('y', t, ensemble_member),
                    2.0, 3.0)
                return [constraint1, constraint2]

        """
        return []

    def path_constraints(self, ensemble_member: int) -> List[
            Tuple[ca.MX, Union[float, np.ndarray], Union[float, np.ndarray]]]:
        """
        Returns a list of path constraints.

        Path constraints apply to all times and ensemble members simultaneously.
        Call :func:`OptimizationProblem.state` to return a time- and ensemble-member-independent
        symbol representing a model variable.

        :param ensemble_member: The ensemble member index. This index may only
                                be used to supply member-dependent bounds.

        :returns: A list of triples ``(f, m, M)``, with an :class:`MX` object representing
                  the path constraint function ``f``, lower bound ``m``, and upper bound ``M``.
                  The bounds may be numbers or :class:`.Timeseries` objects.

        Example::

            def path_constraints(self, ensemble_member):
                # 2 * x must lie between 2 and 4 for every time instance.
                path_constraint1 = (2 * self.state('x'), 2.0, 4.0)
                # x + y must lie between 2 and 3 for every time instance
                path_constraint2 = (self.state('x') + self.state('y'), 2.0, 3.0)
                return [path_constraint1, path_constraint2]

        """
        return []

    def post(self) -> None:
        """
        Postprocessing logic is performed here.
        """
        pass

    @property
    def equidistant(self) -> bool:
        """
        ``True`` if all time series are equidistant.
        """
        return False

    INTERPOLATION_LINEAR = 0
    INTERPOLATION_PIECEWISE_CONSTANT_FORWARD = 1
    INTERPOLATION_PIECEWISE_CONSTANT_BACKWARD = 2

    def interpolate(
            self,
            t: Union[float, np.ndarray],
            ts: np.ndarray,
            fs: np.ndarray,
            f_left: float = np.nan,
            f_right: float = np.nan,
            mode: int = INTERPOLATION_LINEAR) -> Union[float, np.ndarray]:
        """
        Linear interpolation over time.

        :param t:       Time at which to evaluate the interpolant.
        :type t:        float or vector of floats
        :param ts:      Time stamps.
        :type ts:       numpy array
        :param fs:      Function values at time stamps ts.
        :param f_left:  Function value left of leftmost time stamp.
        :param f_right: Function value right of rightmost time stamp.
        :param mode:    Interpolation mode.

        :returns: The interpolated value.
        """

        if isinstance(fs, np.ndarray) and fs.ndim == 2:
            # 2-D array of values. Interpolate each column separately.
            if len(t) == len(ts) and np.all(t == ts):
                # Early termination; nothing to interpolate
                return fs.copy()

            fs_int = [self.interpolate(t, ts, fs[:, i], f_left, f_right, mode) for i in range(fs.shape[1])]
            return np.stack(fs_int, axis=1)
        elif hasattr(t, '__iter__'):
            if len(t) == len(ts) and np.all(t == ts):
                # Early termination; nothing to interpolate
                return fs.copy()

            return self.__interpolate(t, ts, fs, f_left, f_right, mode)
        else:
            if ts[0] == t:
                # Early termination; nothing to interpolate
                return fs[0]

            return self.__interpolate(t, ts, fs, f_left, f_right, mode)

    def __interpolate(self, t, ts, fs, f_left=np.nan, f_right=np.nan, mode=INTERPOLATION_LINEAR):
        """
        Linear interpolation over time.

        :param t:       Time at which to evaluate the interpolant.
        :type t:        float or vector of floats
        :param ts:      Time stamps.
        :type ts:       numpy array
        :param fs:      Function values at time stamps ts.
        :param f_left:  Function value left of leftmost time stamp.
        :param f_right: Function value right of rightmost time stamp.
        :param mode:    Interpolation mode.

        Note that it is assumed that `ts` is sorted. No such assumption is made for `t`
.
        :returns: The interpolated value.
        """

        if f_left is None:
            if (min(t) if hasattr(t, '__iter__') else t) < ts[0]:
                raise Exception("Interpolation: Point {} left of range".format(t))

        if f_right is None:
            if (max(t) if hasattr(t, '__iter__') else t) > ts[-1]:
                raise Exception("Interpolation: Point {} right of range".format(t))

        if mode == self.INTERPOLATION_LINEAR:
            # No need to handle f_left / f_right; NumPy already does this for us
            return np.interp(t, ts, fs, f_left, f_right)
        elif mode == self.INTERPOLATION_PIECEWISE_CONSTANT_FORWARD:
            v = fs[np.maximum(np.searchsorted(ts, t, side='right') - 1, 0)]
        elif mode == self.INTERPOLATION_PIECEWISE_CONSTANT_BACKWARD:
            v = fs[np.minimum(np.searchsorted(ts, t, side='left'), len(ts) - 1)]
        else:
            raise NotImplementedError

        # Handle f_left / f_right:
        if hasattr(t, "__iter__"):
            v[t < ts[0]] = f_left
            v[t > ts[-1]] = f_right
        else:
            if t < ts[0]:
                v = f_left
            elif t > ts[-1]:
                v = f_right

        return v

    @abstractproperty
    def controls(self) -> List[str]:
        """
        List of names of the control variables (excluding aliases).
        """
        pass

    @abstractmethod
    def discretize_controls(self, resolved_bounds: AliasDict) -> Tuple[
            int, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Performs the discretization of the control inputs, filling lower and upper
        bound vectors for the resulting optimization variables, as well as an initial guess.

        :param resolved_bounds: :class:`AliasDict` of numerical bound values. This is the
                                same dictionary as returned by :func:`bounds`, but with all
                                parameter symbols replaced with their numerical values.

        :returns: The number of control variables in the optimization problem, a lower
                  bound vector, an upper bound vector, a seed vector, and a dictionary
                  of offset values.
        """
        pass

    def dynamic_parameters(self) -> List[ca.MX]:
        """
        Returns a list of parameter symbols that may vary from run to run.  The values
        of these parameters are not cached.

        :returns: A list of parameter symbols.
        """
        return []

    @abstractmethod
    def extract_controls(self, ensemble_member: int = 0) -> Dict[str, np.ndarray]:
        """
        Extracts state time series from optimizer results.

        Must return a dictionary of result time series.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of control input time series.
        """
        pass

    def control_vector(self, variable: str, ensemble_member: int = 0) -> Union[ca.MX, List[ca.MX]]:
        """
        Return the optimization variables for the entire time horizon of the given state.

        :param variable:        Variable name.
        :param ensemble_member: The ensemble member index.

        :returns: A vector of control input symbols for the entire time horizon.

        :raises: KeyError
        """
        return self.state_vector(variable, ensemble_member)

    def control(self, variable: str) -> ca.MX:
        """
        Returns an :class:`MX` symbol for the given control input, not bound to any time.

        :param variable: Variable name.

        :returns: :class:`MX` symbol for given control input.

        :raises: KeyError
        """
        return self.variable(variable)

    @abstractmethod
    def control_at(self, variable: str, t: float, ensemble_member: int = 0, scaled: bool = False) -> ca.MX:
        """
        Returns an :class:`MX` symbol representing the given control input at the given time.

        :param variable:        Variable name.
        :param t:               Time.
        :param ensemble_member: The ensemble member index.
        :param scaled:          True to return the scaled variable.

        :returns: :class:`MX` symbol representing the control input at the given time.

        :raises: KeyError
        """
        pass

    @abstractproperty
    def differentiated_states(self) -> List[str]:
        """
        List of names of the differentiated state variables (excluding aliases).
        """
        pass

    @abstractproperty
    def algebraic_states(self) -> List[str]:
        """
        List of names of the algebraic state variables (excluding aliases).
        """
        pass

    @abstractmethod
    def discretize_states(self, resolved_bounds: AliasDict) -> Tuple[
            int, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Perform the discretization of the states.

        Fills lower and upper bound vectors for the resulting optimization
        variables, as well as an initial guess.

        :param resolved_bounds: :class:`AliasDict` of numerical bound values.
            This is the same dictionary as returned by :func:`bounds`, but
            with all parameter symbols replaced with their numerical values.

        :returns: The number of control variables in the optimization problem,
                  a lower bound vector, an upper bound vector, a seed vector,
                  and a dictionary of vector offset values.
        """
        pass

    @abstractmethod
    def extract_states(self, ensemble_member: int = 0) -> Dict[str, np.ndarray]:
        """
        Extracts state time series from optimizer results.

        Must return a dictionary of result time series.

        :param ensemble_member: The ensemble member index.

        :returns: A dictionary of state time series.
        """
        pass

    @abstractmethod
    def state_vector(self, variable: str, ensemble_member: int = 0) -> Union[ca.MX, List[ca.MX]]:
        """
        Return the optimization variables for the entire time horizon of the given state.

        :param variable:        Variable name.
        :param ensemble_member: The ensemble member index.

        :returns: A vector of state symbols for the entire time horizon.

        :raises: KeyError
        """
        pass

    def state(self, variable: str) -> ca.MX:
        """
        Returns an :class:`MX` symbol for the given state, not bound to any time.

        :param variable: Variable name.

        :returns: :class:`MX` symbol for given state.

        :raises: KeyError
        """
        return self.variable(variable)

    @abstractmethod
    def state_at(self, variable: str, t: float, ensemble_member: int = 0, scaled: bool = False) -> ca.MX:
        """
        Returns an :class:`MX` symbol representing the given variable at the given time.

        :param variable:        Variable name.
        :param t:               Time.
        :param ensemble_member: The ensemble member index.
        :param scaled:          True to return the scaled variable.

        :returns: :class:`MX` symbol representing the state at the given time.

        :raises: KeyError
        """
        pass

    @abstractmethod
    def extra_variable(self, variable: str, ensemble_member: int = 0) -> ca.MX:
        """
        Returns an :class:`MX` symbol representing the extra variable inside the state vector.

        :param variable:        Variable name.
        :param ensemble_member: The ensemble member index.

        :returns: :class:`MX` symbol representing the extra variable.

        :raises: KeyError
        """
        pass

    @abstractmethod
    def states_in(self, variable: str, t0: float = None, tf: float = None, ensemble_member: int = 0) -> Iterator[ca.MX]:
        """
        Iterates over symbols for states in the interval [t0, tf].

        :param variable:        Variable name.
        :param t0:              Left bound of interval.  If equal to None, the initial time is used.
        :param tf:              Right bound of interval.  If equal to None, the final time is used.
        :param ensemble_member: The ensemble member index.

        :raises: KeyError
        """
        pass

    @abstractmethod
    def integral(self, variable: str, t0: float = None, tf: float = None, ensemble_member: int = 0) -> ca.MX:
        """
        Returns an expression for the integral over the interval [t0, tf].

        :param variable:        Variable name.
        :param t0:              Left bound of interval.  If equal to None, the initial time is used.
        :param tf:              Right bound of interval.  If equal to None, the final time is used.
        :param ensemble_member: The ensemble member index.

        :returns: :class:`MX` object representing the integral.

        :raises: KeyError
        """
        pass

    @abstractmethod
    def der(self, variable: str) -> ca.MX:
        """
        Returns an :class:`MX` symbol for the time derivative given state, not bound to any time.

        :param variable: Variable name.

        :returns: :class:`MX` symbol for given state.

        :raises: KeyError
        """
        pass

    @abstractmethod
    def der_at(self, variable: str, t: float, ensemble_member: int = 0) -> ca.MX:
        """
        Returns an expression for the time derivative of the specified variable at time t.

        :param variable:        Variable name.
        :param t:               Time.
        :param ensemble_member: The ensemble member index.

        :returns: :class:`MX` object representing the derivative.

        :raises: KeyError
        """
        pass

    def get_timeseries(self, variable: str, ensemble_member: int = 0) -> Timeseries:
        """
        Looks up a timeseries from the internal data store.

        :param variable:        Variable name.
        :param ensemble_member: The ensemble member index.

        :returns: The requested time series.
        :rtype: :class:`.Timeseries`

        :raises: KeyError
        """
        raise NotImplementedError

    def set_timeseries(
            self,
            variable: str,
            timeseries: Timeseries,
            ensemble_member: int = 0,
            output: bool = True,
            check_consistency: bool = True) -> None:
        """
        Sets a timeseries in the internal data store.

        :param variable:          Variable name.
        :param timeseries:        Time series data.
        :type timeseries:         iterable of floats, or :class:`.Timeseries`
        :param ensemble_member:   The ensemble member index.
        :param output:            Whether to include this time series in output data files.
        :param check_consistency: Whether to check consistency between the time stamps on
                                  the new timeseries object and any existing time stamps.
        """
        raise NotImplementedError

    def timeseries_at(self, variable: str, t: float, ensemble_member: int = 0) -> float:
        """
        Return the value of a time series at the given time.

        :param variable:        Variable name.
        :param t:               Time.
        :param ensemble_member: The ensemble member index.

        :returns: The interpolated value of the time series.

        :raises: KeyError
        """
        raise NotImplementedError

    def map_path_expression(self, expr: ca.MX, ensemble_member: int) -> ca.MX:
        """
        Maps the path expression `expr` over the entire time horizon of the optimization problem.

        :param expr: An :class:`MX` path expression.

        :returns: An :class:`MX` expression evaluating `expr` over the entire time horizon.
        """
        raise NotImplementedError

    @debug_check(DebugLevel.HIGH)
    def __debug_check_linearity_constraints(self, nlp):
        x = nlp['x']
        f = nlp['f']
        g = nlp['g']

        expand_f_g = ca.Function('f_g', [x], [f, g]).expand()
        X_sx = ca.SX.sym('X', *x.shape)
        f_sx, g_sx = expand_f_g(X_sx)

        jac = ca.Function('j', [X_sx], [ca.jacobian(g_sx, X_sx)]).expand()
        if jac(np.nan).is_regular():
            logger.info("The constraints are linear")
        else:
            hes = ca.Function('j', [X_sx], [ca.jacobian(ca.jacobian(g_sx, X_sx), X_sx)]).expand()
            if hes(np.nan).is_regular():
                logger.info("The constraints are quadratic")
            else:
                logger.info("The constraints are nonlinear")

    @debug_check(DebugLevel.VERYHIGH)
    def __debug_check_linear_independence(self, lbx, ubx, lbg, ubg, nlp):
        x = nlp['x']
        f = nlp['f']
        g = nlp['g']

        expand_f_g = ca.Function('f_g', [x], [f, g]).expand()
        x_sx = ca.SX.sym('X', *x.shape)
        f_sx, g_sx = expand_f_g(x_sx)

        x, f, g = x_sx, f_sx, g_sx

        lbg = np.array(ca.vertsplit(ca.veccat(*lbg))).ravel()
        ubg = np.array(ca.vertsplit(ca.veccat(*ubg))).ravel()

        # Find the linear constraints
        g_sjac = ca.Function('Af', [x], [ca.jtimes(g, x, x.ones(*x.shape))])

        res = g_sjac(np.nan)
        res = np.array(res).ravel()
        g_is_linear = ~np.isnan(res)

        # Find the rows in the jacobian with only a single entry
        g_jac_csr = ca.DM(ca.Function('tmp', [x], [g]).sparsity_jac(0, 0)).tocsc().tocsr()
        g_single_variable = (np.diff(g_jac_csr.indptr) == 1)

        # Find the rows which are equality constraints
        g_eq_constraint = (lbg == ubg)

        # The intersection of all selections are constraints like we want
        g_constant_assignment = g_is_linear & g_single_variable & g_eq_constraint

        # Map of variable (index) to constraints/row numbers
        var_index_assignment = {}
        for i in range(g.size1()):
            if g_constant_assignment[i]:
                var_ind = g_jac_csr.getrow(i).indices[0]
                var_index_assignment.setdefault(var_ind, []).append(i)

        var_names, named_x, named_f, named_g = self._debug_get_named_nlp(nlp)

        for vi, g_inds in var_index_assignment.items():
            if len(g_inds) > 1:
                logger.info("Variable '{}' has duplicate constraints setting its value:".format(var_names[vi]))
                for g_i in g_inds:
                    logger.info("row {}: {} = {}".format(g_i, named_g[g_i], lbg[g_i]))

        # Find variables for which the bounds are equal, but also an equality
        # constraint is set. This would result in a constraint `1 = 1` with
        # the default IPOPT option `fixed_variable_treatment = make_parameter`
        x_inds = np.flatnonzero(lbx == ubx)

        for vi in x_inds:
            if vi in var_index_assignment:
                logger.info("Variable '{}' has equal bounds (value = {}), but also the following equality constraints:"
                            .format(var_names[vi], lbx[vi]))
                for g_i in var_index_assignment[vi]:
                    logger.info("row {}: {} = {}".format(g_i, named_g[g_i], lbg[g_i]))



import bisect
import logging
from abc import ABCMeta, abstractmethod
from math import isfinite

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools.simulation.simulation_problem import SimulationProblem

logger = logging.getLogger("rtctools")


class IOMixin(SimulationProblem, metaclass=ABCMeta):
    """
    Base class for all IO methods of optimization problems.
    """

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        self._simulation_times = []

        self.__first_update_call = True

    def pre(self) -> None:
        # Call read method to read all input
        self.read()

        self._simulation_times = []

    @abstractmethod
    def read(self) -> None:
        """
        Reads input data from files, storing it in the internal data store through the various set or add methods
        """
        pass

    def post(self) -> None:
        # Call write method to write all output
        self.write()

    @abstractmethod
    def write(self) -> None:
        """"
        Writes output data to files, getting the data from the data store through the various get methods
        """
        pass

    def initialize(self, config_file=None):
        # Set up experiment
        timeseries_import_times = self.io.times_sec
        self.__dt = timeseries_import_times[1] - timeseries_import_times[0]
        self.setup_experiment(0, timeseries_import_times[-1], self.__dt)

        parameter_variables = set(self.get_parameter_variables())

        logger.debug("Model parameters are {}".format(parameter_variables))

        for parameter, value in self.io.parameters().items():
            if parameter in parameter_variables:
                logger.debug("IOMixin: Setting parameter {} = {}".format(parameter, value))
                self.set_var(parameter, value)

        # Load input variable names
        self.__input_variables = set(self.get_input_variables().keys())

        # Set input values
        t_idx = bisect.bisect_left(timeseries_import_times, 0.0)
        self.__set_input_variables(t_idx)

        logger.debug("Model inputs are {}".format(self.__input_variables))

        # Set first timestep
        self._simulation_times.append(self.get_current_time())

        # Empty output
        self._io_output_variables = self.get_output_variables()
        self._io_output = AliasDict(self.alias_relation)

        # Call super, which will also initialize the model itself
        super().initialize(config_file)

        # Extract consistent t0 values
        for variable in self._io_output_variables:
            self._io_output[variable] = [self.get_var(variable)]

    def __set_input_variables(self, t_idx, use_cache=False):
        if not use_cache:
            self.__cache_loop_timeseries = {}

            timeseries_names = set(self.io.get_timeseries_names(0))
            for v in self.get_variables():
                if v in timeseries_names:
                    _, values = self.io.get_timeseries_sec(v)
                    self.__cache_loop_timeseries[v] = values

        for variable, values in self.__cache_loop_timeseries.items():
            value = values[t_idx]
            if isfinite(value):
                self.set_var(variable, value)
            else:
                logger.debug("IOMixin: Found bad value {} at index [{}] in timeseries aliased to input {}"
                             .format(value, t_idx, variable))

    def update(self, dt):
        # Time step
        if dt < 0:
            dt = self.__dt

        # Current time stamp
        t = self.get_current_time()
        self._simulation_times.append(t + dt)

        # Get current time index
        t_idx = bisect.bisect_left(self.io.times_sec, t + dt)

        # Set input values
        self.__set_input_variables(t_idx, not self.__first_update_call)

        # Call super
        super().update(dt)

        # Extract results
        for variable, values in self._io_output.items():
            values.append(self.get_var(variable))

        self.__first_update_call = False

    @cached
    def parameters(self):
        """
        Return a dictionary of parameters, including parameters in the input files files.

        :returns: Dictionary of parameters
        """
        # Call parent class first for default values.
        parameters = super().parameters()

        # Load parameters from input files (stored in internal data store)
        for parameter_name, value in self.io.parameters().items():
            parameters[parameter_name] = value

        if logger.getEffectiveLevel() == logging.DEBUG:
            for parameter_name in self.io.parameters().keys():
                logger.debug("IOMixin: Read parameter {}".format(parameter_name))

        return parameters

    def times(self, variable=None):
        """
        Return a list of all the timesteps in seconds.

        :param variable: Variable name.

        :returns: List of all the timesteps in seconds.
        """
        idx = bisect.bisect_left(self.io.datetimes, self.io.reference_datetime)
        return self.io.times_sec[idx:]

    def timeseries_at(self, variable, t):
        """
        Return the value of a time series at the given time.

        :param variable: Variable name.
        :param t: Time.

        :returns: The interpolated value of the time series.

        :raises: KeyError
        """
        timeseries_times_sec, values = self.io.get_timeseries_sec(variable)
        t_idx = bisect.bisect_left(timeseries_times_sec, t)
        if timeseries_times_sec[t_idx] == t:
            return values[t_idx]
        else:
            return np.interp(t, timeseries_times_sec, values)



from typing import Dict

from .optimization_problem import OptimizationProblem


class LinearizationMixin(OptimizationProblem):
    """
    Adds linearized equation parameter bookkeeping to your optimization aproblem.

    If your model contains linearized equations, this mixin will set the
    parameters of these equations based on the t0 value of an associated
    timeseries.

    The mapping between linearization parameters and time series is provided
    in the ``linearization_parameters`` method.
    """

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)

        for parameter, timeseries_id in self.linearization_parameters().items():
            parameters[parameter] = self.timeseries_at(timeseries_id, self.initial_time, ensemble_member)

        return parameters

    def linearization_parameters(self) -> Dict[str, str]:
        """
        :returns: A dictionary of parameter names mapping to time series identifiers.
        """

        return {}



import logging
from typing import Dict, Union

from .optimization_problem import OptimizationProblem
from .timeseries import Timeseries

logger = logging.getLogger("rtctools")


class HomotopyMixin(OptimizationProblem):
    """
    Adds homotopy to your optimization problem.  A homotopy is a continuous transformation between
    two optimization problems, parametrized by a single parameter :math:`\\theta \\in [0, 1]`.

    Homotopy may be used to solve non-convex optimization problems, by starting with a convex
    approximation at :math:`\\theta = 0.0` and ending with the non-convex problem at
    :math:`\\theta = 1.0`.

    .. note::

        It is advised to look for convex reformulations of your problem, before resorting to a use
        of the (potentially expensive) homotopy process.

    """

    def seed(self, ensemble_member):
        seed = super().seed(ensemble_member)
        options = self.homotopy_options()

        # Overwrite the seed only when the results of the latest run are
        # stored within this class. That is, when the GoalProgrammingMixin
        # class is not used or at the first run of the goal programming loop.
        if self.__theta > options['theta_start'] and getattr(self, '_gp_first_run', True):
            for key, result in self.__results[ensemble_member].items():
                times = self.times(key)
                if ((result.ndim == 1 and len(result) == len(times))
                        or (result.ndim == 2 and result.shape[0] == len(times))):
                    # Only include seed timeseries which are consistent
                    # with the specified time stamps.
                    seed[key] = Timeseries(times, result)
                elif ((result.ndim == 1 and len(result) == 1)
                        or (result.ndim == 2 and result.shape[0] == 1)):
                    seed[key] = result
        return seed

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)

        options = self.homotopy_options()
        try:
            # Only set the theta if we are in the optimization loop. We want
            # to avoid accidental usage of the parameter value in e.g. pre().
            # Note that we use a try-except here instead of hasattr, to avoid
            # explicit name mangling.
            parameters[options['homotopy_parameter']] = self.__theta
        except AttributeError:
            pass

        return parameters

    def homotopy_options(self) -> Dict[str, Union[str, float]]:
        """
        Returns a dictionary of options controlling the homotopy process.

        +------------------------+------------+---------------+
        | Option                 | Type       | Default value |
        +========================+============+===============+
        | ``theta_start``        | ``float``  | ``0.0``       |
        +------------------------+------------+---------------+
        | ``delta_theta_0``      | ``float``  | ``1.0``       |
        +------------------------+------------+---------------+
        | ``delta_theta_min``    | ``float``  | ``0.01``      |
        +------------------------+------------+---------------+
        | ``homotopy_parameter`` | ``string`` | ``theta``     |
        +------------------------+------------+---------------+

        The homotopy process is controlled by the homotopy parameter in the model, specified
        by the option ``homotopy_parameter``.  The homotopy parameter is initialized to ``theta_start``,
        and increases to a value of ``1.0`` with a dynamically changing step size.  This step
        size is initialized with the value of the option ``delta_theta_0``.  If this step
        size is too large, i.e., if the problem with the increased homotopy parameter fails to
        converge, the step size is halved.  The process of halving terminates when the step size falls
        below the minimum value specified by the option ``delta_theta_min``.

        :returns: A dictionary of homotopy options.
        """

        return {'theta_start': 0.0,
                'delta_theta_0': 1.0,
                'delta_theta_min': 0.01,
                'homotopy_parameter': 'theta'}

    def dynamic_parameters(self):
        dynamic_parameters = super().dynamic_parameters()

        if self.__theta > 0:
            # For theta = 0, we don't mark the homotopy parameter as being dynamic,
            # so that the correct sparsity structure is obtained for the linear model.
            options = self.homotopy_options()
            dynamic_parameters.append(self.variable(options['homotopy_parameter']))

        return dynamic_parameters

    def optimize(self, preprocessing=True, postprocessing=True, log_solver_failure_as_error=True):
        # Pre-processing
        if preprocessing:
            self.pre()

        options = self.homotopy_options()
        delta_theta = options['delta_theta_0']

        # Homotopy loop
        self.__theta = options['theta_start']

        while self.__theta <= 1.0:
            logger.info("Solving with homotopy parameter theta = {}.".format(self.__theta))

            success = super().optimize(preprocessing=False, postprocessing=False, log_solver_failure_as_error=False)
            if success:
                self.__results = [
                    self.extract_results(ensemble_member) for ensemble_member in range(self.ensemble_size)]

                if self.__theta == 0.0:
                    self.check_collocation_linearity = False
                    self.linear_collocation = False

                    # Recompute the sparsity structure for the nonlinear model family.
                    self.clear_transcription_cache()

            else:
                if self.__theta == options['theta_start']:
                    break

                self.__theta -= delta_theta
                delta_theta /= 2

                if delta_theta < options['delta_theta_min']:
                    failure_message = (
                        'Solver failed with homotopy parameter theta = {}. Theta cannot '
                        'be decreased further, as that would violate the minimum delta '
                        'theta of {}.'.format(self.__theta, options['delta_theta_min']))
                    if log_solver_failure_as_error:
                        logger.error(failure_message)
                    else:
                        # In this case we expect some higher level process to deal
                        # with the solver failure, so we only log it as info here.
                        logger.info(failure_message)
                    break

            self.__theta += delta_theta

        # Post-processing
        if postprocessing:
            self.post()

        return success



import casadi as ca

import numpy as np

from rtctools.optimization.goal_programming_mixin_base import Goal, StateGoal, _GoalProgrammingMixinBase
from rtctools.optimization.goal_programming_mixin_base import _GoalConstraint


class LinearizedOrderGoal(Goal):
    #: Override linearization of goal order. Related global goal programming
    #: option is ``linearize_goal_order`` (see :py:meth:`LinearizedOrderGoalProgrammingMixin.goal_programming_options`).
    #: The default value of None defers to the global option, but the user can
    #: explicitly override it per goal by setting this value to True or False.
    linearize_order = None

    #: Coefficients to linearize a goal's order
    _linear_coefficients = {}

    @classmethod
    def _get_linear_coefficients(cls, order, eps=0.1, kind='balanced'):
        assert order > 1, "Order should be strictly larger than one"

        try:
            return cls._linear_coefficients[eps][order]
        except KeyError:
            pass

        x = ca.SX.sym('x')
        a = ca.SX.sym('a')
        b = ca.SX.sym('b')

        # Strike a balance between "absolute error < eps" and "relative error < eps" by
        # multiplying eps with x**(order-1)
        if kind == 'balanced':
            f = x**order - eps * x**(order-1) - (a * x + b)
        elif kind == 'abs':
            f = x**order - eps - (a * x + b)
        else:
            raise Exception("Unknown error approximation strategy '{}'".format(kind))

        res_vals = ca.Function("res_vals", [x, ca.vertcat(a, b)], [f])

        do_step = ca.rootfinder("next_state", "fast_newton", res_vals)

        x = 0.0
        a = 0.0
        b = 0.0

        xs = [0.0]
        while x < 1.0:
            # Initial guess larger than 1.0 to always have the next point be
            # on the right (i.e. not left) side.
            x = float(do_step(2.0, [a, b]))
            a = order * x ** (order - 1)
            b = x**order - a * x
            xs.append(x)

        # Turn underestimate into an overestimate, such that we get rid of
        # horizontal line at origin.
        xs[-1] = 1.0
        xs = np.array(xs)
        ys = xs**order

        a = (ys[1:] - ys[:-1])/(xs[1:] - xs[:-1])
        b = ys[1:] - a * xs[1:]
        lines = list(zip(a, b))

        cls._linear_coefficients.setdefault(eps, {})[order] = lines

        return lines


class LinearizedOrderStateGoal(LinearizedOrderGoal, StateGoal):
    """
    Convenience class definition for linearized order state goals. Note that
    it is possible to just inherit from :py:class:`.LinearizedOrderGoal` to get the needed
    functionality for control of the linearization at goal level.
    """
    pass


class LinearizedOrderGoalProgrammingMixin(_GoalProgrammingMixinBase):
    """
    Adds support for linearization of the goal objective functions, i.e. the
    violation variables to a certain power. This can be used to keep a problem
    fully linear and/or make sure that no quadratic constraints appear when using
    the goal programming option ``keep_soft_constraints``.
    """

    def goal_programming_options(self):
        """
        If ``linearize_goal_order`` is set to ``True``, the goal's order will be
        approximated linearly for any goals where order > 1. Note that this option
        does not work with minimization goals of higher order. Instead, it is
        suggested to transform these minimization goals into goals with a target (and
        function range) when using this option. Note that this option can be overriden
        on the level of a goal by using a :py:class:`LinearizedOrderGoal` (see
        :py:attr:`LinearizedOrderGoal.linearize_order`).
        """
        options = super().goal_programming_options()
        options['linearize_goal_order'] = True
        return options

    def _gp_validate_goals(self, goals, is_path_goal):
        options = self.goal_programming_options()

        for goal in goals:
            goal_linearize = None
            if isinstance(goal, LinearizedOrderGoal):
                goal_linearize = goal.linearize_order

            if goal_linearize or (options['linearize_goal_order'] and goal_linearize is not False):
                if not goal.has_target_bounds and goal.order > 1:
                    raise Exception("Higher order minimization goals not allowed with "
                                    "`linearize_goal_order` for goal {}".format(goal))

        super()._gp_validate_goals(goals, is_path_goal)

    def _gp_goal_constraints(self, goals, sym_index, options, is_path_goal):
        options = self.goal_programming_options()

        def _linearize_goal(goal):
            goal_linearize = None
            if isinstance(goal, LinearizedOrderGoal):
                goal_linearize = goal.linearize_order

            if goal_linearize or (options['linearize_goal_order'] and goal_linearize is not False):
                if goal.order > 1 and not goal.critical:
                    return True
                else:
                    return False
            else:
                return False

        lo_soft_constraints = [[] for ensemble_member in range(self.ensemble_size)]
        lo_epsilons = []

        # For the linearized goals, we use all of the normal processing,
        # except for the objective. We can override the objective function by
        # setting a _objective_func function on the Goal object.
        for j, goal in enumerate(goals):
            if not _linearize_goal(goal):
                continue

            assert goal.has_target_bounds, "Cannot linearize minimization goals"

            # Make a linear epsilon, and constraints relating the linear
            # variable to the original objective function
            path_prefix = "path_" if is_path_goal else ""
            linear_variable = ca.MX.sym(path_prefix + "lineps_{}_{}".format(sym_index, j), goal.size)

            lo_epsilons.append(linear_variable)

            if isinstance(goal, LinearizedOrderGoal):
                coeffs = goal._get_linear_coefficients(goal.order)
            else:
                coeffs = LinearizedOrderGoal._get_linear_coefficients(goal.order)

            epsilon_name = path_prefix + "eps_{}_{}".format(sym_index, j)

            for a, b in coeffs:
                # We add to soft constraints, as these constraints are no longer valid when
                # having `keep_soft_constraints` = False. This is because the `epsilon` and
                # the `linear_variable` no longer exist in the next priority.
                for ensemble_member in range(self.ensemble_size):
                    def _f(problem,
                           goal=goal, epsilon_name=epsilon_name, linear_variable=linear_variable, a=a, b=b,
                           ensemble_member=ensemble_member, is_path_constraint=is_path_goal):
                        if is_path_constraint:
                            eps = problem.variable(epsilon_name)
                            lin = problem.variable(linear_variable.name())
                        else:
                            eps = problem.extra_variable(epsilon_name, ensemble_member)
                            lin = problem.extra_variable(linear_variable.name(), ensemble_member)

                        return lin - a * eps - b

                    lo_soft_constraints[ensemble_member].append(_GoalConstraint(goal, _f, 0.0, np.inf, False))

            if is_path_goal and options['scale_by_problem_size']:
                goal_m, goal_M = self._gp_min_max_arrays(goal, target_shape=len(self.times()))
                goal_active = np.isfinite(goal_m) | np.isfinite(goal_M)
                n_active = np.sum(goal_active.astype(int), axis=0)
            else:
                n_active = 1

            def _objective_func(problem, ensemble_member,
                                goal=goal, linear_variable=linear_variable, is_path_goal=is_path_goal,
                                n_active=n_active):
                if is_path_goal:
                    lin = problem.variable(linear_variable.name())
                else:
                    lin = problem.extra_variable(linear_variable.name(), ensemble_member)

                return goal.weight * lin / n_active

            goal._objective_func = _objective_func

        epsilons, objectives, soft_constraints, hard_constraints, extra_constants = \
            super()._gp_goal_constraints(goals, sym_index, options, is_path_goal)

        epsilons = epsilons + lo_epsilons
        for ensemble_member in range(self.ensemble_size):
            soft_constraints[ensemble_member].extend(lo_soft_constraints[ensemble_member])

        return epsilons, objectives, soft_constraints, hard_constraints, extra_constants



import configparser
import glob
import logging
import os
import pickle
from typing import Iterable, List, Tuple, Union

import casadi as ca

import numpy as np

import rtctools.data.csv as csv
from rtctools._internal.caching import cached
from rtctools.data.interpolation.bspline1d import BSpline1D
from rtctools.data.interpolation.bspline2d import BSpline2D
from rtctools.optimization.timeseries import Timeseries

from scipy.interpolate import bisplev, bisplrep, splev
from scipy.optimize import brentq

from .optimization_problem import LookupTable as LookupTableBase
from .optimization_problem import OptimizationProblem

logger = logging.getLogger("rtctools")


class LookupTable(LookupTableBase):
    """
    Lookup table.
    """

    def __init__(self, inputs: List[ca.MX], function: ca.Function, tck: Tuple = None):
        """
        Create a new lookup table object.

        :param inputs: List of lookup table input variables.
        :param function: Lookup table CasADi :class:`Function`.
        """
        self.__inputs = inputs
        self.__function = function

        self.__t, self.__c, self.__k = [None] * 3

        if tck is not None:
            if len(tck) == 3:
                self.__t, self.__c, self.__k = tck
            elif len(tck) == 5:
                self.__t = tck[:2]
                self.__c = tck[2]
                self.__k = tck[3:]

    @property
    @cached
    def domain(self) -> Tuple:
        t = self.__t
        if t is None:
            raise AttributeError('This lookup table was not instantiated with tck metadata. \
                                  Domain/Range information is unavailable.')
        if type(t) == tuple and len(t) == 2:
            raise NotImplementedError('Domain/Range information is not yet implemented for 2D LookupTables')

        return np.nextafter(t[0], np.inf), np.nextafter(t[-1], -np.inf)

    @property
    @cached
    def range(self) -> Tuple:
        return self(self.domain[0]), self(self.domain[1])

    @property
    def inputs(self) -> List[ca.MX]:
        """
        List of lookup table input variables.
        """
        return self.__inputs

    @property
    def function(self) -> ca.Function:
        """
        Lookup table CasADi :class:`Function`.
        """
        return self.__function

    @property
    @cached
    def __numeric_function_evaluator(self):
        return np.vectorize(
            lambda *args: np.nan
            if np.any(np.isnan(args))
            else np.float(self.function(*args))
        )

    def __call__(
        self, *args: Union[float, Iterable, Timeseries]
    ) -> Union[float, np.ndarray, Timeseries]:
        """
        Evaluate the lookup table.

        :param args: Input values.
        :type args: Float, iterable of floats, or :class:`.Timeseries`
        :returns: Lookup table evaluated at input values.

        Example use::

            y = lookup_table(1.0)
            [y1, y2] = lookup_table([1.0, 2.0])

        """
        evaluator = self.__numeric_function_evaluator
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, Timeseries):
                return Timeseries(arg.times, self(arg.values))
            else:
                if hasattr(arg, "__iter__"):
                    arg = np.fromiter(arg, dtype=float)
                    return evaluator(arg)
                else:
                    arg = float(arg)
                    return evaluator(arg).item()
        else:
            if any(isinstance(arg, Timeseries) for arg in args):
                raise TypeError(
                    "Higher-order LookupTable calls do not yet support Timeseries parameters"
                )
            elif any(hasattr(arg, "__iter__") for arg in args):
                raise TypeError(
                    "Higher-order LookupTable calls do not yet support vector parameters"
                )
            else:
                args = np.fromiter(args, dtype=float)
                return evaluator(*args)

    def reverse_call(
        self,
        y: Union[float, Iterable, Timeseries],
        domain: Tuple[float, float] = (None, None),
        detect_range_error: bool = True,
    ) -> Union[float, np.ndarray, Timeseries]:
        """Do an inverted call on this LookupTable

        Uses SciPy brentq optimizer to simulate a reversed call.
        Note: Method does not work with higher-order LookupTables
        """
        if isinstance(y, Timeseries):
            # Recurse and return
            return Timeseries(y.times, self.reverse_call(y.values))

        # Get domain information
        l_d, u_d = domain
        if l_d is None:
            l_d = self.domain[0]
        if u_d is None:
            u_d = self.domain[1]

        # Cast y to array of float
        if hasattr(y, "__iter__"):
            y_array = np.fromiter(y, dtype=float)
        else:
            y_array = np.array([y], dtype=float)

        # Find not np.nan
        is_not_nan = ~np.isnan(y_array)
        y_array_not_nan = y_array[is_not_nan]

        # Detect if there is a range violation
        if detect_range_error:
            l_r, u_r = self.range
            lb_viol = y_array_not_nan < l_r
            ub_viol = y_array_not_nan > u_r
            all_viol = y_array_not_nan[lb_viol | ub_viol]
            if all_viol.size > 0:
                raise ValueError(
                    "Values {} are not in lookup table range ({}, {})".format(
                        all_viol, l_r, u_r
                    )
                )

        # Construct function to do inverse evaluation
        evaluator = self.__numeric_function_evaluator

        def inv_evaluator(y_target):
            """inverse evaluator function"""
            return brentq(lambda x: evaluator(x) - y_target, l_d, u_d)

        inv_evaluator = np.vectorize(inv_evaluator)

        # Calculate x_array
        x_array = np.full_like(y_array, np.nan, dtype=float)
        if y_array_not_nan.size != 0:
            x_array[is_not_nan] = inv_evaluator(y_array_not_nan)

        # Return x
        if hasattr(y, "__iter__"):
            return x_array
        else:
            return x_array.item()


class CSVLookupTableMixin(OptimizationProblem):
    """
    Adds lookup tables to your optimization problem.

    During preprocessing, the CSV files located inside the ``lookup_tables`` subfolder are read.
    In every CSV file, the first column contains the output of the lookup table.  Subsequent columns contain
    the input variables.

    Cubic B-Splines are used to turn the data points into continuous lookup tables.

    Optionally, a file ``curvefit_options.ini`` may be included inside the ``lookup_tables`` folder.
    This file contains, grouped per lookup table, the following options:

    * monotonicity:
        * is an integer, magnitude is ignored
        * if positive, causes spline to be monotonically increasing
        * if negative, causes spline to be monotonically decreasing
        * if 0, leaves spline monotonicity unconstrained

    * curvature:
        * is an integer, magnitude is ignored
        * if positive, causes spline curvature to be positive (convex)
        * if negative, causes spline curvature to be negative (concave)
        * if 0, leaves spline curvature unconstrained

    .. note::

        Currently only one-dimensional lookup tables are fully supported.  Support for two-
        dimensional lookup tables is experimental.

    :cvar csv_delimiter:                 Column delimiter used in CSV files.  Default is ``,``.
    :cvar csv_lookup_table_debug:        Whether to generate plots of the spline fits.  Default is ``false``.
    :cvar csv_lookup_table_debug_points: Number of evaluation points for plots.  Default is ``100``.
    """

    #: Column delimiter used in CSV files
    csv_delimiter = ','

    #: Debug settings
    csv_lookup_table_debug = False
    csv_lookup_table_debug_points = 100

    def __init__(self, **kwargs):
        # Check arguments
        if 'input_folder' in kwargs:
            assert('lookup_table_folder' not in kwargs)

            self.__lookup_table_folder = os.path.join(
                kwargs['input_folder'], 'lookup_tables')
        else:
            self.__lookup_table_folder = kwargs['lookup_table_folder']

        # Call parent
        super().__init__(**kwargs)

    def pre(self):
        # Call parent class first for default behaviour.
        super().pre()

        # Get curve fitting options from curvefit_options.ini file
        ini_path = os.path.join(
            self.__lookup_table_folder, 'curvefit_options.ini')
        try:
            ini_config = configparser.RawConfigParser()
            ini_config.read(ini_path)
            no_curvefit_options = False
        except IOError:
            logger.info(
                "CSVLookupTableMixin: No curvefit_options.ini file found. Using default values.")
            no_curvefit_options = True

        def get_curvefit_options(curve_name, no_curvefit_options=no_curvefit_options):
            if no_curvefit_options:
                return 0, 0, 0

            curvefit_options = []

            def get_property(prop_name):
                try:
                    prop = int(ini_config.get(curve_name, prop_name))
                except configparser.NoSectionError:
                    prop = 0
                except configparser.NoOptionError:
                    prop = 0
                except ValueError:
                    raise Exception(
                        'CSVLookupTableMixin: Invalid {0} constraint for {1}. {0} should '
                        'be either -1, 0, or 1.'.format(prop_name, curve_name))
                return prop

            for prop_name in ['monotonicity', 'monotonicity2', 'curvature']:
                curvefit_options.append(get_property(prop_name))

            logger.debug("CSVLookupTableMixin: Curve fit option for {}:({},{},{})".format(
                curve_name, *curvefit_options))
            return tuple(curvefit_options)

        # Read CSV files
        logger.info(
            "CSVLookupTableMixin: Generating Splines from lookup table data.")
        self.__lookup_tables = {}
        for filename in glob.glob(os.path.join(self.__lookup_table_folder, "*.csv")):

            logger.debug(
                "CSVLookupTableMixin: Reading lookup table from {}".format(filename))

            csvinput = csv.load(filename, delimiter=self.csv_delimiter)
            output = csvinput.dtype.names[0]
            inputs = csvinput.dtype.names[1:]

            # Get monotonicity and curvature from ini file
            mono, mono2, curv = get_curvefit_options(output)

            logger.debug(
                "CSVLookupTableMixin: Output is {}, inputs are {}.".format(output, inputs))

            tck = None
            function = None

            # If tck file is newer than the csv file, first try to load the cached values from the tck file
            tck_filename = filename.replace('.csv', '.tck')
            valid_cache = False
            if os.path.exists(tck_filename):
                if no_curvefit_options:
                    valid_cache = os.path.getmtime(filename) < os.path.getmtime(tck_filename)
                else:
                    valid_cache = (os.path.getmtime(filename) < os.path.getmtime(tck_filename)) and \
                                  (os.path.getmtime(ini_path) < os.path.getmtime(tck_filename))
                if valid_cache:
                    logger.debug(
                        'CSVLookupTableMixin: Attempting to use cached tck values for {}'.format(output))
                    with open(tck_filename, 'rb') as f:
                        try:
                            tck, function = pickle.load(f)
                        except Exception:
                            valid_cache = False
            if not valid_cache:
                logger.info(
                    'CSVLookupTableMixin: Recalculating tck values for {}'.format(output))

            if len(csvinput.dtype.names) == 2:
                if not valid_cache:
                    k = 3  # default value
                    # 1D spline fitting needs k+1 data points
                    if len(csvinput[output]) >= k + 1:
                        tck = BSpline1D.fit(csvinput[inputs[0]], csvinput[
                                            output], k=k, monotonicity=mono, curvature=curv)
                    else:
                        raise Exception(
                            'CSVLookupTableMixin: Too few data points in {} to do spline fitting. '
                            'Need at least {} points.'.format(filename, k + 1))

                if self.csv_lookup_table_debug:
                    import pylab
                    i = np.linspace(csvinput[inputs[0]][0], csvinput[
                                    inputs[0]][-1], self.csv_lookup_table_debug_points)
                    o = splev(i, tck)
                    pylab.clf()
                    # TODO: Figure out why cross-section B0607 in NZV does not
                    # conform to constraints!
                    pylab.plot(i, o)
                    pylab.plot(csvinput[inputs[0]], csvinput[
                               output], linestyle='', marker='x', markersize=10)
                    figure_filename = filename.replace('.csv', '.png')
                    pylab.savefig(figure_filename)

                symbols = [ca.SX.sym(inputs[0])]
                if not valid_cache:
                    function = ca.Function('f', symbols, [BSpline1D(*tck)(symbols[0])])
                self.__lookup_tables[output] = LookupTable(symbols, function, tck)

            elif len(csvinput.dtype.names) == 3:
                if tck is None:
                    kx = 3  # default value
                    ky = 3  # default value

                    # 2D spline fitting needs (kx+1)*(ky+1) data points
                    if len(csvinput[output]) >= (kx + 1) * (ky + 1):
                        # TODO: add curvature parameters from curvefit_options.ini
                        # once 2d fit is implemented
                        tck = bisplrep(csvinput[inputs[0]], csvinput[
                                       inputs[1]], csvinput[output], kx=kx, ky=ky)
                    else:
                        raise Exception(
                            'CSVLookupTableMixin: Too few data points in {} to do spline fitting. '
                            'Need at least {} points.'.format(filename, (kx + 1) * (ky + 1)))

                if self.csv_lookup_table_debug:
                    import pylab
                    i1 = np.linspace(csvinput[inputs[0]][0], csvinput[
                                     inputs[0]][-1], self.csv_lookup_table_debug_points)
                    i2 = np.linspace(csvinput[inputs[1]][0], csvinput[
                                     inputs[1]][-1], self.csv_lookup_table_debug_points)
                    i1, i2 = np.meshgrid(i1, i2)
                    i1 = i1.flatten()
                    i2 = i2.flatten()
                    o = bisplev(i1, i2, tck)
                    pylab.clf()
                    pylab.plot_surface(i1, i2, o)
                    figure_filename = filename.replace('.csv', '.png')
                    pylab.savefig(figure_filename)
                symbols = [ca.SX.sym(inputs[0]), ca.SX.sym(inputs[1])]
                if not valid_cache:
                    function = ca.Function('f', symbols, [BSpline2D(*tck)(symbols[0], symbols[1])])
                self.__lookup_tables[output] = LookupTable(symbols, function, tck)

            else:
                raise Exception(
                    'CSVLookupTableMixin: {}-dimensional lookup tables not implemented yet.'.format(
                        len(csvinput.dtype.names)))

            if not valid_cache:
                pickle.dump((tck, function),
                            open(filename.replace('.csv', '.tck'), 'wb'),
                            protocol=pickle.HIGHEST_PROTOCOL)

    def lookup_tables(self, ensemble_member):
        # Call parent class first for default values.
        lookup_tables = super().lookup_tables(ensemble_member)

        # Update lookup_tables with imported csv lookup tables
        lookup_tables.update(self.__lookup_tables)

        return lookup_tables



import logging
import os
import xml.etree.ElementTree as ET
from collections import namedtuple

ts_ids = namedtuple('ids', 'location_id parameter_id qualifier_id')
p_ids = namedtuple('ids', 'model_id location_id parameter_id')

ns = {'fews': 'http://www.wldelft.nl/fews',
      'pi': 'http://www.wldelft.nl/fews/PI'}

logger = logging.getLogger("rtctools")


class DataConfig:
    """
    rtcDataConfig wrapper.

    Used to map PI timeseries to RTC-Tools variable names.
    """

    def __init__(self, folder):
        """
        Parse rtcDataConfig file

        :param folder: Folder in which rtcDataConfig.xml is located.
        """
        self.__variable_map = {}
        self.__location_parameter_ids = {}
        self.__parameter_map = {}
        self.__model_parameter_ids = {}

        path = os.path.join(folder, "rtcDataConfig.xml")
        try:
            tree = ET.parse(path)
            root = tree.getroot()

            timeseriess1 = root.findall('./*/fews:timeSeries', ns)
            timeseriess2 = root.findall('./fews:timeSeries', ns)
            timeseriess1.extend(timeseriess2)

            for timeseries in timeseriess1:
                pi_timeseries = timeseries.find('fews:PITimeSeries', ns)
                if pi_timeseries is not None:
                    internal_id = timeseries.get('id')
                    external_id = self.__pi_timeseries_id(pi_timeseries, 'fews')

                    if internal_id in self.__location_parameter_ids:
                        message = (
                            'Found more than one external timeseries '
                            'mapped to internal id {} in {}.').format(internal_id, path)
                        logger.error(message)
                        raise Exception(message)
                    elif external_id in self.__variable_map:
                        message = (
                            'Found more than one internal timeseries '
                            'mapped to external id {} in {}.').format(external_id, path)
                        logger.error(message)
                        raise Exception(message)
                    else:
                        self.__location_parameter_ids[internal_id] = \
                            self.__pi_location_parameter_id(pi_timeseries, 'fews')
                        self.__variable_map[external_id] = internal_id

            for k in ['import', 'export']:
                res = root.find(
                    './fews:%s/fews:PITimeSeriesFile/fews:timeSeriesFile' % k, ns)
                if res is not None:
                    setattr(self, 'basename_%s' %
                            k, os.path.splitext(res.text)[0])

            parameters = root.findall('./fews:parameter', ns)
            if parameters is not None:
                for parameter in parameters:
                    pi_parameter = parameter.find('fews:PIParameter', ns)
                    if pi_parameter is not None:
                        internal_id = parameter.get('id')
                        external_id = self.__pi_parameter_id(pi_parameter, 'fews')

                        if internal_id in self.__model_parameter_ids:
                            message = (
                                'Found more than one external parameter mapped '
                                'to internal id {} in {}.').format(internal_id, path)
                            logger.error(message)
                            raise Exception(message)
                        if external_id in self.__parameter_map:
                            message = (
                                'Found more than one interal parameter mapped to external '
                                'modelId {}, locationId {}, parameterId {} in {}.').format(
                                external_id.model_id, external_id.location_id, external_id.parameter_id, path)
                            logger.error(message)
                            raise Exception(message)
                        else:
                            self.__model_parameter_ids[internal_id] = self.__pi_model_parameter_id(pi_parameter, 'fews')
                            self.__parameter_map[external_id] = internal_id

        except IOError:
            logger.error(
                'No rtcDataConfig.xml file was found in "{}".'.format(folder))
            raise

    def __pi_timeseries_id(self, el, namespace):
        location_id = el.find(namespace + ':locationId', ns).text
        parameter_id = el.find(namespace + ':parameterId', ns).text

        timeseries_id = location_id + ':' + parameter_id

        qualifiers = el.findall(namespace + ':qualifierId', ns)
        qualifier_ids = []
        for qualifier in qualifiers:
            qualifier_ids.append(qualifier.text)

        if len(qualifier_ids) > 0:
            qualifier_ids.sort()

            return timeseries_id + ':' + ':'.join(qualifier_ids)
        else:
            return timeseries_id

    def __pi_location_parameter_id(self, el, namespace):
        qualifier_ids = []
        qualifiers = el.findall(namespace + ':qualifierId', ns)
        for qualifier in qualifiers:
            qualifier_ids.append(qualifier.text)

        location_parameter_ids = ts_ids(location_id=el.find(namespace + ':locationId', ns).text,
                                        parameter_id=el.find(namespace + ':parameterId', ns).text,
                                        qualifier_id=qualifier_ids)
        return location_parameter_ids

    def __pi_parameter_id(self, el, namespace):
        model_id = el.find(namespace + ':modelId', ns).text
        location_id = el.find(namespace + ':locationId', ns).text
        parameter_id = el.find(namespace + ':parameterId', ns).text

        return self.__long_parameter_id(parameter_id, location_id, model_id)

    def __pi_model_parameter_id(self, el, namespace):
        model_id = el.find(namespace + ':modelId', ns).text
        location_id = el.find(namespace + ':locationId', ns).text
        parameter_id = el.find(namespace + ':parameterId', ns).text

        model_parameter_ids = p_ids(model_id=(model_id if model_id is not None else ""),
                                    location_id=(location_id if location_id is not None else ""),
                                    parameter_id=(parameter_id if parameter_id is not None else ""))

        return model_parameter_ids

    def __long_parameter_id(self, parameter_id, location_id=None, model_id=None):
        """
        Convert a model, location and parameter combination to a single parameter id
        of the form model:location:parameter.
        """
        if location_id is not None:
            parameter_id = location_id + ':' + parameter_id
        if model_id is not None:
            parameter_id = model_id + ':' + parameter_id
        return parameter_id

    def variable(self, pi_header):
        """
        Map a PI timeseries header to an RTC-Tools timeseries ID.

        :param pi_header: XML ElementTree node containing a PI timeseries header.

        :returns: A timeseries ID.
        :rtype: string
        """
        series_id = self.__pi_timeseries_id(pi_header, 'pi')
        try:
            return self.__variable_map[series_id]
        except KeyError:
            return series_id

    def pi_variable_ids(self, variable):
        """
        Map an RTC-Tools timeseries ID to a named tuple of location, parameter
        and qualifier ID's.

        :param variable: A timeseries ID.

        :returns: A named tuple with fields location_id, parameter_id and qualifier_id.
        :rtype: namedtuple
        :raises KeyError: If the timeseries ID has no mapping in rtcDataConfig.
        """
        return self.__location_parameter_ids[variable]

    def parameter(self, parameter_id, location_id=None, model_id=None):
        """
        Map a combination of parameter ID, location ID, model ID to an
        RTC-Tools parameter ID.

        :param parameter_id: String with parameter ID
        :param location_id: String with location ID
        :param model_id: String with model ID

        :returns: A parameter ID.
        :rtype: string
        :raises KeyError: If the combination has no mapping in rtcDataConfig.
        """
        parameter_id_long = self.__long_parameter_id(parameter_id, location_id, model_id)

        return self.__parameter_map[parameter_id_long]

    def pi_parameter_ids(self, parameter):
        """
        Map an RTC-Tools model parameter ID to a named tuple of model, location
        and parameter ID's.

        :param parameter: A model parameter ID.

        :returns: A named tuple with fields model_id, location_id and parameter_id.
        :rtype: namedtuple
        :raises KeyError: If the parameter ID has no mapping in rtcDataConfig.
        """
        return self.__model_parameter_ids[parameter]



import logging
from datetime import timedelta

import rtctools.data.pi as pi
import rtctools.data.rtc as rtc
from rtctools.optimization.io_mixin import IOMixin

logger = logging.getLogger("rtctools")


class PIMixin(IOMixin):
    """
    Adds `Delft-FEWS Published Interface
    <https://publicwiki.deltares.nl/display/FEWSDOC/The+Delft-Fews+Published+Interface>`_
    I/O to your optimization problem.

    During preprocessing, files named ``rtcDataConfig.xml``, ``timeseries_import.xml``, ``rtcParameterConfig.xml``,
    and ``rtcParameterConfig_Numerical.xml`` are read from the ``input`` subfolder.  ``rtcDataConfig.xml`` maps
    tuples of FEWS identifiers, including location and parameter ID, to RTC-Tools time series identifiers.

    During postprocessing, a file named ``timeseries_export.xml`` is written to the ``output`` subfolder.

    :cvar pi_binary_timeseries:
        Whether to use PI binary timeseries format. Default is ``False``.
    :cvar pi_parameter_config_basenames:
        List of parameter config file basenames to read. Default is [``rtcParameterConfig``].
    :cvar pi_parameter_config_numerical_basename:
        Numerical config file basename to read. Default is ``rtcParameterConfig_Numerical``.
    :cvar pi_check_for_duplicate_parameters:
        Check if duplicate parameters are read. Default is ``True``.
    :cvar pi_validate_timeseries:
        Check consistency of timeseries. Default is ``True``.
    """

    #: Whether to use PI binary timeseries format
    pi_binary_timeseries = False

    #: Location of rtcParameterConfig files
    pi_parameter_config_basenames = ['rtcParameterConfig']
    pi_parameter_config_numerical_basename = 'rtcParameterConfig_Numerical'

    #: Check consistency of timeseries
    pi_validate_timeseries = True

    #: Check for duplicate parameters
    pi_check_for_duplicate_parameters = True

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        # Load rtcDataConfig.xml.  We assume this file does not change over the
        # life time of this object.
        self.__data_config = rtc.DataConfig(self._input_folder)

    def read(self):
        # Call parent class first for default behaviour.
        super().read()

        # rtcParameterConfig
        self.__parameter_config = []
        try:
            for pi_parameter_config_basename in self.pi_parameter_config_basenames:
                self.__parameter_config.append(pi.ParameterConfig(
                    self._input_folder, pi_parameter_config_basename))
        except IOError:
            raise Exception(
                "PIMixin: {}.xml not found in {}.".format(pi_parameter_config_basename, self._input_folder))

        try:
            self.__parameter_config_numerical = pi.ParameterConfig(
                  self._input_folder, self.pi_parameter_config_numerical_basename)
        except IOError:
            self.__parameter_config_numerical = None

        try:
            self.__timeseries_import = pi.Timeseries(
                self.__data_config, self._input_folder, self.timeseries_import_basename,
                binary=self.pi_binary_timeseries, pi_validate_times=self.pi_validate_timeseries)
        except IOError:
            raise Exception("PIMixin: {}.xml not found in {}.".format(
                self.timeseries_import_basename, self._input_folder))

        self.__timeseries_export = pi.Timeseries(
            self.__data_config, self._output_folder, self.timeseries_export_basename,
            binary=self.pi_binary_timeseries, pi_validate_times=False, make_new_file=True)

        # Convert timeseries timestamps to seconds since t0 for internal use
        timeseries_import_times = self.__timeseries_import.times

        # Timestamp check
        if self.pi_validate_timeseries:
            for i in range(len(timeseries_import_times) - 1):
                if timeseries_import_times[i] >= timeseries_import_times[i + 1]:
                    raise Exception(
                        'PIMixin: Time stamps must be strictly increasing.')

        if self.__timeseries_import.dt:
            # Check if the timeseries are truly equidistant
            if self.pi_validate_timeseries:
                dt = timeseries_import_times[1] - timeseries_import_times[0]
                for i in range(len(timeseries_import_times) - 1):
                    if timeseries_import_times[i + 1] - timeseries_import_times[i] != dt:
                        raise Exception(
                            'PIMixin: Expecting equidistant timeseries, the time step '
                            'towards {} is not the same as the time step(s) before. Set '
                            'unit to nonequidistant if this is intended.'.format(
                                timeseries_import_times[i + 1]))

        # Offer input timeseries to IOMixin
        self.io.reference_datetime = self.__timeseries_import.forecast_datetime

        for ensemble_member in range(self.__timeseries_import.ensemble_size):
            for variable, values in self.__timeseries_import.items(ensemble_member):
                self.io.set_timeseries(variable, timeseries_import_times, values, ensemble_member)

            # store the parameters in the internal data store. Note that we
            # are effectively broadcasting parameters, as ParameterConfig does
            # not support parameters varying per ensemble member
            for parameter_config in self.__parameter_config:
                for location_id, model_id, parameter_id, value in parameter_config:
                    try:
                        parameter = self.__data_config.parameter(parameter_id, location_id, model_id)
                    except KeyError:
                        parameter = parameter_id

                    self.io.set_parameter(parameter, value, ensemble_member,
                                          check_duplicates=self.pi_check_for_duplicate_parameters)

    def solver_options(self):
        # Call parent
        options = super().solver_options()

        # Only do this if we have a rtcParameterConfig_Numerical
        if self.__parameter_config_numerical is None:
            return options

        # Load solver options from parameter config
        for _location_id, _model, option, value in self.__parameter_config_numerical:
            options[option] = value

        # Done
        return options

    def write(self):
        # Call parent class first for default behaviour.
        super().write()

        # Get time stamps
        times = self.times()
        if len(set(times[1:] - times[:-1])) == 1:
            dt = timedelta(seconds=times[1] - times[0])
        else:
            dt = None

        # Start of write output
        # Write the time range for the export file.
        self.__timeseries_export.times = [
            self.__timeseries_import.times[self.__timeseries_import.forecast_index]
            + timedelta(seconds=s) for s in times]

        # Write other time settings
        self.__timeseries_export.forecast_datetime = self.__timeseries_import.forecast_datetime
        self.__timeseries_export.dt = dt
        self.__timeseries_export.timezone = self.__timeseries_import.timezone

        # Write the ensemble properties for the export file.
        if self.ensemble_size > 1:
            self.__timeseries_export.contains_ensemble = True
        self.__timeseries_export.ensemble_size = self.ensemble_size
        self.__timeseries_export.contains_ensemble = (self.ensemble_size > 1)

        # Start looping over the ensembles for extraction of the output values.
        for ensemble_member in range(self.ensemble_size):
            results = self.extract_results(ensemble_member)

            # For all variables that are output variables the values are
            # extracted from the results.
            for variable in [sym.name() for sym in self.output_variables]:
                try:
                    values = results[variable]
                    if len(values) != len(times):
                        values = self.interpolate(
                            times, self.times(variable), values, self.interpolation_method(variable))
                except KeyError:
                    try:
                        ts = self.get_timeseries(variable, ensemble_member)
                        if len(ts.times) != len(times):
                            values = self.interpolate(
                                times, ts.times, ts.values)
                        else:
                            values = ts.values
                    except KeyError:
                        logger.error(
                            'PIMixin: Output requested for non-existent variable {}. '
                            'Will not be in output file.'.format(variable))
                        continue

                # Check if ID mapping is present
                try:
                    self.__data_config.pi_variable_ids(variable)
                except KeyError:
                    logger.debug(
                        'PIMixin: variable {} has no mapping defined in rtcDataConfig '
                        'so cannot be added to the output file.'.format(variable))
                    continue

                # Add series to output file.
                # NOTE: We use the unit of the zeroth ensemble member, as
                # we might be outputting more ensembles than we read in.
                self.__timeseries_export.set(
                    variable, values,
                    unit=self.__timeseries_import.get_unit(variable, ensemble_member=0),
                    ensemble_member=ensemble_member)

        # Write output file to disk
        self.__timeseries_export.write()

    def set_timeseries(self, variable: str, *args, unit: str = None, **kwargs):
        if unit is not None:
            self.__timeseries_import.set_unit(variable, unit, 0)

        super().set_timeseries(variable, *args, **kwargs)

    @property
    def timeseries_import(self):
        """
        :class:`pi.Timeseries` object containing the input data.
        """
        return self.__timeseries_import

    @property
    def timeseries_import_times(self):
        """
        List of time stamps for which input data is specified.

        The time stamps are in seconds since t0, and may be negative.
        """
        return self.io.times_sec

    @property
    def timeseries_export(self):
        """
        :class:`pi.Timeseries` object for holding the output data.
        """
        return self.__timeseries_export
